import re

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
types = parameters.get('types')

for resource in datapackage['resources']:

    name = resource['name']
    if not resources.match(name):
        continue

    fields = resource.setdefault('schema', {}).get('fields', [])
    for field_name, field_definition in types.items():
        field_name_re = re.compile(field_name)
        if field_definition is not None:
            filtered_fields = list(
                filter(
                    lambda f: field_name_re.fullmatch(f['name']) is not None,  # pylint: disable=cell-var-from-loop
                    fields
                )
            )
            for field in filtered_fields:
                field.update(field_definition)
            assert len(filtered_fields) > 0, \
                "No field found matching %r" % field_name
        else:
            fields = list(
                filter(
                    lambda f: field_name_re.fullmatch(f['name']) is None,  # pylint: disable=cell-var-from-loop
                    fields
                )
            )

    resource['schema']['fields'] = fields

spew(datapackage, resource_iterator)



from .celery_app import celery_app

from ..manager.tasks import execute_pipeline
from ..manager.status import status

@celery_app.task
def execute_pipeline_task(pipeline_id,
                          pipeline_steps,
                          pipeline_cwd,
                          trigger,
                          queue_time):

    last_queue_time = status.get_status(pipeline_id).get('queued')
    if queue_time in {0, last_queue_time}:
        execute_pipeline(pipeline_id,
                         pipeline_steps,
                         pipeline_cwd,
                         trigger,
                         False)



import os
import subprocess
import sys
import json

ROOT_PATH = os.path.join(os.path.dirname(__file__), '..', '..')
ENV = os.environ.copy()
ENV['PYTHONPATH'] = ROOT_PATH


def _test_single_fixture(processor, parameters, dp_in, data_in, dp_out, data_out):
    """Test a single processor with fixture"""

    process = subprocess.run([sys.executable, processor, '1', parameters, 'False', ''],
                             input=data_in, stdout=subprocess.PIPE, env=ENV)
    output = process.stdout.decode('utf8')
    (actual_dp, *actual_data) = output.split('\n\n', 1)
    assert actual_dp == dp_out
    if len(actual_data) > 0:
        actual_data = actual_data[0]
        for actual, expected in zip(actual_data.split('\n'), data_out.split('\n')):
            if len(actual) == 0:
                assert len(expected) == 0
            else:
                assert rejsonize(actual) == rejsonize(expected)


def rejsonize(s):
    return json.dumps(json.loads(s), sort_keys=True, ensure_ascii=True)

prerecorded_responses = \
    os.walk(
        os.path.join(
            os.path.dirname(__file__),
            'fixtures'
        )
    )
for dirpath, _, filenames in prerecorded_responses:
    for filename in filenames:
        parts = \
            open(os.path.join(dirpath, filename), encoding='utf8').read().split('--\n')
        # print("PPP %s" % '\n~~~\n'.join(parts))
        (processor, params, dp_in, data_in, dp_out, data_out) = parts
        processor = processor.replace('.', '/')
        processor = os.path.join(ROOT_PATH, 'datapackage_pipelines', 'lib', processor.strip()+'.py')
        params = rejsonize(params)
        dp_out = rejsonize(dp_out)
        dp_in = rejsonize(dp_in)
        data_in = (dp_in + '\n\n' + data_in).encode('utf8')

        def inner(processor_, params_, dp_in_, data_in_, dp_out_, data_out_):
            def inner2():
                return _test_single_fixture(processor_, params_, dp_in_, data_in_, dp_out_, data_out_)
            return inner2
        globals()['test_stdlib_%s' % filename] = inner(processor, params, dp_in, data_in, dp_out, data_out)



import logging
import os

PROCESSOR_PATH = os.environ.get('DPP_PROCESSOR_PATH', '').split(';')


def find_file_in_path(path, remove=0):
    def finder(parts):
        filename = os.path.join(*(path + parts[remove:]))
        if os.path.exists(filename):
            return filename
    return finder


def convert_dot_notation(executor):
    parts = []
    back_up = False
    while executor.startswith('..'):
        parts.append('..')
        executor = executor[1:]
        back_up = True

    if executor.startswith('.'):
        executor = executor[1:]

    executor = executor.split('.')
    executor[-1] += '.py'

    parts.extend(executor)

    return back_up, parts


def load_module(module):
    module_name = 'datapackage_pipelines_'+module
    try:
        module = __import__(module_name)
        return module
    except ImportError:
        pass


def resolve_executor(executor, path):

    back_up, parts = convert_dot_notation(executor)
    resolvers = [find_file_in_path([path])]
    if not back_up:
        if len(parts) > 1:
            module_name = parts[0]
            module = load_module(module_name)
            if module is not None:
                module = list(module.__path__)[0]
                resolvers.append(find_file_in_path([module, 'processors'], 1))

        resolvers.extend([
            find_file_in_path([path])
            for path in PROCESSOR_PATH
        ])
        resolvers.append(find_file_in_path([os.path.dirname(__file__),
                                            '..', 'lib']))

    for resolver in resolvers:
        location = resolver(parts)
        if location is not None:
            return location

    raise FileNotFoundError("Couldn't resolve {0} at {1}"
                            .format(executor, path))

resolved_generators = {}


def resolve_generator(module_name):
    if module_name in resolved_generators:
        return resolved_generators[module_name]
    resolved_generators[module_name] = None
    module = load_module(module_name)
    if module is None:
        return None
    try:
        generator_class = module.Generator
    except AttributeError:
        logging.warning("Can't find 'Generator' identifier in %s", module_name)
        return None
    generator = generator_class()
    resolved_generators[module_name] = generator
    return generator



import os
from jsontableschema_sql import Storage
from sqlalchemy import create_engine

from datapackage_pipelines.lib.dump.dumper_base import DumperBase


class SQLDumper(DumperBase):

    def initialize(self, parameters):
        table_to_resource = parameters['tables']
        engine = parameters.get('engine', 'env://DPP_DB_ENGINE')
        if engine.startswith('env://'):
            engine = os.environ.get(engine[6:])
            assert engine is not None
        self.engine = create_engine(engine)  # pylint: disable=attribute-defined-outside-init

        for k, v in table_to_resource.items():
            v['table-name'] = k

        # pylint: disable=attribute-defined-outside-init
        self.converted_resources = \
            dict((v['resource-name'], v) for v in table_to_resource.values())

    def handle_resource(self, resource, spec, parameters, datapackage):
        resource_name = spec['name']
        if resource_name not in self.converted_resources:
            return resource
        else:
            converted_resource = self.converted_resources[resource_name]
            table_name = converted_resource['table-name']
            storage = Storage(self.engine, prefix=table_name)
            if '' in storage.buckets:
                storage.delete('')
            storage.create('', spec['schema'])
            return storage.write('', resource, keyed=True, as_generator=True)

SQLDumper()()



import logging
import click

from .manager import execute_pipeline, finalize
from .manager.status import status
from .manager.specs import pipelines


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        click.echo('Available Pipelines:')
        for pipeline_id, _, _, dirty, errors in pipelines():
            click.echo('- {} {}{}'
                       .format(pipeline_id,
                               '(*)' if dirty else '',
                               '(E)' if len(errors) > 0 else ''))
            for short, long in errors:
                click.echo('\t{}: {}'.format(short, long))


@cli.command()
def serve():
    """Start the web server"""
    from .web import app
    app.run(host='0.0.0.0', debug=True, port=5000)


@cli.command()
@click.argument('pipeline_id')
@click.option('--use-cache/--no-use-cache', default=True)
def run(pipeline_id, use_cache):
    """Run a pipeline by pipeline-id.
Use 'all' for running all pipelines,
or 'dirty' for running just the dirty ones."""
    try:
        results = []
        for _pipeline_id, pipeline_details, pipeline_cwd, dirty, errors \
                in pipelines():
            if len(errors) != 0:
                continue
            if ((_pipeline_id == pipeline_id) or
                    (pipeline_id == 'all') or
                    (pipeline_id == 'dirty' and dirty)):
                success, stats = \
                    execute_pipeline(_pipeline_id,
                                     pipeline_details.get('pipeline', []),
                                     pipeline_cwd,
                                     use_cache=use_cache)
                results.append((_pipeline_id, success, stats))
                if _pipeline_id == pipeline_id:
                    break

        logging.info('RESULTS:')
        for pipeline_id, success, stats in results:
            logging.info('%s: %s %s',
                         'SUCCESS' if success else 'FAILURE',
                         pipeline_id,
                         repr(stats) if stats is not None else '')

    except KeyboardInterrupt:
        pass
    finally:
        finalize()


@cli.command()
def init():
    """Reset the status of all pipelines"""
    status.initialize()



import os
import hashlib

import yaml

from ..utilities.extended_json import json
from .resolver import resolve_executor, resolve_generator
from .status import status

SPEC_FILENAME = 'pipeline-spec.yaml'
SOURCE_FILENAME_SUFFIX = '.source-spec.yaml'


def find_pipeline_specs(dirpath, filenames):
    if SPEC_FILENAME in filenames:
        abspath = os.path.abspath(dirpath)
        fullpath = os.path.join(abspath, SPEC_FILENAME)
        try:
            with open(fullpath, encoding='utf8') as spec_file:
                spec = yaml.load(spec_file.read())
                for pipeline_id, pipeline_details in spec.items():
                    pipeline_id = os.path.join(dirpath, pipeline_id)
                    yield abspath, pipeline_id, pipeline_details, None, None
        except yaml.YAMLError as e:
            yield abspath, dirpath, {}, None, ('Invalid Pipeline Spec', str(e))


# pylint: disable=too-many-locals
def find_source_specs(dirpath, filenames):
    abspath = os.path.abspath(dirpath)
    for filename in filenames:
        if not filename.endswith(SOURCE_FILENAME_SUFFIX):
            continue

        fullpath = os.path.join(abspath, filename)

        module_name = filename[:-len(SOURCE_FILENAME_SUFFIX)]
        generator = resolve_generator(module_name)
        if generator is None:
            message = 'Unknown source description kind "{}" in {}'\
                      .format(module_name, fullpath)
            yield abspath, dirpath, {}, None, ('Unknown source kind', message)
            continue

        try:
            with open(fullpath, encoding='utf8') as spec_file:
                source_spec = yaml.load(spec_file.read())
                if generator.internal_validate(source_spec):
                    spec = generator.internal_generate(source_spec)
                    for pipeline_id, schedule, steps in spec:
                        pipeline_details = {
                            'schedule': {'crontab': schedule},
                            'pipeline': steps
                        }
                        pipeline_id = os.path.join(dirpath, pipeline_id)
                        yield abspath, pipeline_id, \
                              pipeline_details, source_spec, None
                else:
                    message = 'Invalid source description for "{}" in {}'\
                              .format(module_name, fullpath)
                    yield abspath, dirpath, {}, \
                          None, ('Invalid Source', message)
        except yaml.YAMLError as e:
            yield abspath, dirpath, {}, None, ('Invalid Source Spec', str(e))


def find_specs(root_dir='.'):
    for dirpath, _, filenames in os.walk(root_dir):

        # Pipeline Specs
        for p in find_pipeline_specs(dirpath, filenames):
            yield p

        # Source specs:
        for p in find_source_specs(dirpath, filenames):
            yield p


def validate_required_keys(obj, keys, abspath, errors):
    for required_key in keys:
        if required_key not in obj:
            message = 'Missing parameter {0} in {1}'\
                      .format(required_key, abspath)
            errors.append(('Missing Parameter', message))


# pylint: disable=too-many-locals
def validate_specs():

    all_pipeline_ids = set()

    for abspath, pipeline_id, \
            pipeline_details, source_details, error in find_specs():

        errors = []
        if error is not None:
            errors.append(error)

        if pipeline_id in all_pipeline_ids:
            message = 'Duplicate key {0} in {1}'\
                      .format(pipeline_id, abspath)
            errors.append(('Duplicate Pipeline Id', message))

        validate_required_keys(pipeline_details,
                               ['pipeline'],
                               abspath,
                               errors)

        pipeline = pipeline_details.get('pipeline', [])

        try:
            for step in pipeline:
                validate_required_keys(step, ['run'], abspath, errors)
                executor = step['run']
                step['name'] = executor
                step['run'] = resolve_executor(executor, abspath)
        except FileNotFoundError as e:
            errors.append(('Unresolved processor', str(e)))

        cache_hash = ''
        if len(errors) == 0:
            for step in pipeline:
                m = hashlib.md5()
                m.update(cache_hash.encode('ascii'))
                m.update(open(step['run'], 'rb').read())
                m.update(json.dumps(step, ensure_ascii=True, sort_keys=True)
                         .encode('ascii'))
                cache_hash = m.hexdigest()
                step['_cache_hash'] = cache_hash

        schedule = pipeline_details.get('schedule', {})
        if 'crontab' in schedule:
            schedule = schedule['crontab'].split()
            pipeline_details['schedule'] = schedule

        dirty = status.register(pipeline_id, cache_hash,
                                pipeline=pipeline_details,
                                source=source_details,
                                errors=errors)

        yield pipeline_id, pipeline_details, abspath, dirty, errors


def pipelines():
    return validate_specs()



import json
import sqlite3
import tempfile
import collections

from datapackage_pipelines.wrapper import ingest, spew


# pylint: disable=too-few-public-methods
class KeyCalc(object):

    def __init__(self, key_spec):
        if isinstance(key_spec, list):
            key_spec = ':'.join('{%s}' % key for key in key_spec)
        self.key_spec = key_spec

    def __call__(self, row):
        return self.key_spec.format(**row)


class DB(object):

    def __init__(self):
        self.tmpfile = tempfile.NamedTemporaryFile()
        self.db = sqlite3.connect(self.tmpfile.name)
        self.cursor = self.db.cursor()
        self.cursor.execute('''CREATE TABLE d (key text, value text)''')
        self.cursor.execute('''CREATE UNIQUE INDEX i ON d (key)''')

    def get(self, key):
        ret = self.cursor.execute('''SELECT value FROM d WHERE key=?''',
                                  (key,)).fetchone()
        if ret is None:
            return None
        else:
            return json.loads(ret[0])

    def set(self, key, value):
        value = json.dumps(value)
        if self.get(key) is not None:
            self.cursor.execute('''UPDATE d SET value=? WHERE key=?''',
                                (value, key))
        else:
            self.cursor.execute('''INSERT INTO d VALUES (?, ?)''',
                                (key, value))
        self.db.commit()

db = DB()


def identity(x):
    return x

Aggregator = collections.namedtuple('Aggregator',
                                    ['func', 'finaliser', 'dataType'])
AGGREGATORS = {
    'sum': Aggregator(lambda curr, new:
                      new + curr if curr is not None else new,
                      identity,
                      None),
    'avg': Aggregator(lambda curr, new:
                      (curr[0] + 1, new + curr[1])
                      if curr is not None
                      else (1, new),
                      lambda value: value[1] / value[0],
                      None),
    'max': Aggregator(lambda curr, new:
                      max(new, curr) if curr is not None else new,
                      identity,
                      None),
    'min': Aggregator(lambda curr, new:
                      min(new, curr) if curr is not None else new,
                      identity,
                      None),
    'first': Aggregator(lambda curr, new:
                        curr if curr is not None else new,
                        identity,
                        None),
    'last': Aggregator(lambda curr, new: new,
                       identity,
                       None),
    'count': Aggregator(lambda curr, new:
                        curr+1 if curr is not None else 1,
                        identity,
                        'number'),
    'any': Aggregator(lambda curr, new: new,
                      identity,
                      None),
}

parameters, datapackage, resource_iterator = ingest()

source = parameters['source']
source_name = source['name']
source_key = KeyCalc(source['key'])
source_delete = source.get('delete', False)

target = parameters['target']
target_name = target['name']
target_key = KeyCalc(target['key'])

fields = parameters['fields']
full = parameters.get('full', True)


def get_aggregator(field):
    spec = fields[field]
    agg = spec.get('aggregate', 'any')
    return agg


def indexer(resource):
    for row in resource:
        key = source_key(row)
        current = db.get(key)
        if current is None:
            current = {}
        for field, spec in fields.items():
            curr = current.get(field)
            agg = get_aggregator(field)
            if agg != 'count':
                new = row.get(spec['name'])
            else:
                new = None
            current[field] = AGGREGATORS[agg].func(curr, new)
        db.set(key, current)
        yield row


def process_target(resource):
    empty_extra = dict((f, None) for f in fields.keys())
    for row in resource:
        key = target_key(row)
        extra = db.get(key)
        if extra is None:
            if not full:
                continue
            extra = empty_extra
        extra = dict(
            (k, AGGREGATORS[get_aggregator(k)].finaliser(v))
            for k, v in extra.items()
        )
        row.update(extra)
        yield row


def new_resource_iterator(resource_iterator_):
    has_index = False
    for resource in resource_iterator_:
        name = resource.spec['name']
        if name == source_name:
            has_index = True
            if source_delete:
                # just empty the iterable
                collections.deque(indexer(resource), maxlen=0)
            else:
                yield indexer(resource)
        elif name == target_name:
            assert has_index
            yield process_target(resource)
        else:
            yield resource


def process_datapackage(datapackage_):

    new_resources = []
    source_spec = None
    for resource in datapackage_['resources']:

        if resource['name'] == source_name:
            source_spec = resource
            if source_delete:
                continue

        elif resource['name'] == target_name:
            assert isinstance(source_spec, dict), \
                "Source resource must appear before target resource"
            target_fields = resource['schema']['fields']
            added_fields = sorted(fields.keys())
            for field in added_fields:
                spec = fields[field]
                agg = get_aggregator(field)
                data_type = AGGREGATORS[agg].dataType
                if data_type is None:
                    source_field = \
                        next(filter(lambda f, spec_=spec:
                                    f['name'] == spec_['name'],
                                    source_spec['schema']['fields']))  # pylint: disable=unsubscriptable-object
                    data_type = source_field['type']
                target_fields.append({
                    'name': field,
                    'type': data_type
                })

        new_resources.append(resource)

    datapackage_['resources'] = new_resources
    return datapackage_


spew(process_datapackage(datapackage),
     new_resource_iterator(resource_iterator))
